/*
   ESP32_SD_Mic_RS.ino
   Audio Rain Sensor.
   Can also be used to back test pre-recorded wav files
   // updated to ratio 8, mag 4000 after adding rain gauge to side and replacing screws with SS
   
   Using ESP32 board support V3.0.3
   compiling for Dfrobot Firebeetle 2 ESP32-E
*/
/*
   (c)2024 Forward Computing and Control Pty. Ltd.
   NSW Australia, www.forward.com.au
   This code is not warranted to be fit for any purpose. You may only use it at your own risk.
   This generated code may be freely used for both private and commercial use
   provided this copyright is maintained.
*/

#include "ESP_I2S.h"
#include "FS.h"
#include "SD.h"
#include "SPI.h"

#include "parseWAV.h"
#include "testWAVinput.h"
#include "LogFile.h"

#include "BufferedOutput.h"
#include "Goertzel.h"
#include "writeWAV.h"
#include <HS_AsyncTCP.h>
#include <WiFi.h>
#include "millisDelay.h"
#include "ProcessBuffers.h"
#include "VolatileVars.h" // control and data vars
#include "loopTimer.h"
#include "ESP32_WS2812Flasher.h"
#include "ESPAutoWiFiConfig.h"
#include "ESPBufferedClient.h"

// uncomment this to log delay processing of microphone recordings
//#define LOG_FILE

// for normal running comment out the #define TESTING ... below
// for testing use either
// #define TESTING BACK_TESTING to test tuning settings only logs detected rain pulse and start/stop rain triggers to file log.txt
// or
// #define TESTING DUMP_ALL_RATIOS_AND_MAGNITUDES   to log the ratio and magnitudes for every 0.1sec with 0.05sec overlap file mags.txt
// these are used below do not change
#define BACK_TESTING 2
#define DUMP_ALL_RATIOS_AND_MAGNITUDES 3
// -----------------------------------

// uncomment one if these two defines to test pre-recorded wav files
//#define TESTING BACK_TESTING
//#define TESTING DUMP_ALL_RATIOS_AND_MAGNITUDES

// TESTING does not connect to WiFi, otherwise tries to connect to WiFi.

//================ DO NOT CHANGE THESE DEFINES use TESTING define above
#ifdef TESTING
#define LOG_FILE

#if TESTING == DUMP_ALL_RATIOS_AND_MAGNITUDES
// this outputs the ratios and magnitudes for every interval. Intervals are 0.1sec with 0.05sec overlaps to file mags.txt
#define MAGNITUDE_FILE
#else
// uncomment this to output the detected rain drops and the start and end of isRaining to log.txt file
#define LOG_RAIN_PULSES
#endif // TESTING == DUMP_ALL_RATIOS_AND_MAGNITUDES
#endif //TESTING

// network settings set by ESPAutoWiFiConfig

static const int portNo = 23; // What port to listen on for connections. telnet
static const unsigned long CONNECTION_TIMEOUT_MS = 10000; // 0 => never time out, else close connection if no data received for this time (ms)
// see asyncConnectionTimeout(CONNECTION_TIMEOUT_MS); in startTelentServer()

#ifdef LOG_RAIN_PULSES
bool logRainPulses = true; // set this to true to log detected rain pulses and print to Serial
#else
bool logRainPulses = false; // set this to true to log detected rain pulses and print to Serial
#endif

//#define DEBUG
Print *debugPtr = NULL;

createBufferedOutput(bufferedOut, 1024, DROP_IF_FULL, false);

#ifdef LOG_FILE
const char logFilename[] = "/log.txt";//The name of the logFile
#else
const char *logFilename = NULL;
#endif

cSF(sfLogLine, 200); // for logging

#ifdef MAGNITUDE_FILE
const char filenameMag[] = "/mags.txt";//The name of the saved Mag file
#else
const char *filenameMag = NULL; // if you don't want the Magitude and ratios file
#endif


uint32_t deltaT_counter_ignore = 10; // skip first 0.5 sec read from mics
const float printScaleFactor = 1e-5; // scales output prints by this number as print(float) prints ovf for numbers > 4294967040

// for i2s mics             // Dfrobot Firebeetle 2 ESP32-E pins
const uint8_t I2S_SCK = 25; // D2   (SCK)
const uint8_t I2S_WS = 16;  // D11  (WS)
const uint8_t I2S_DIN = 26; // D3  (DO)
// const uint8_t IS2_MODE = 4; //  (SEL) Not connected, hard wired as below
// Inside Mic (LEFT) has SEL connected to GND
// Outside Mic (RIGHT) has SEL connected to 3V3

// for SD card  NOTE: do not use D9 (GPI02) as that interfers with reprogramming
const int SD_CS_pin = 13; // D7; //SD cs pin


#define SAMPLE_FREQ (16000)
const int16_t DOWNSAMPLED_FREQ = SAMPLE_FREQ / 4;
#define DATA_BIT_WIDTH (I2S_DATA_BIT_WIDTH_32BIT)
const size_t bytesPerSample = (DATA_BIT_WIDTH / 8);

const size_t rec_buffer_sampleSize = SAMPLE_FREQ / 20 * 2; // number of bytes for 800 samples 0.05sec x 16000Hz X 2 for stereo
const size_t rec_buffer_byteSize = rec_buffer_sampleSize * bytesPerSample; // for stereo
char rec_buffer_1[rec_buffer_byteSize];  // note char buffer of size rec_buffer_byteSize
char rec_buffer_2[rec_buffer_byteSize];  // note char buffer of size rec_buffer_byteSize

// these hold the mono signals derived from rec_buffers (or read in from pre-recorded wave forms)
const size_t mono_buffer_byteSize = rec_buffer_byteSize / 2; // mono buffers
const size_t mono_buffer_sampleSize = mono_buffer_byteSize / bytesPerSample;
char inside_buffer[mono_buffer_byteSize];  // note char buffer of mono of rec_buffer
char outside_buffer[mono_buffer_byteSize];  // note char buffer of mono of rec_buffer

const size_t processBuffer_sampleSize = mono_buffer_sampleSize * 2 / 4 ; // number of bytes for 400 samples 0.1sec downsampled for FT
int32_t processBufferInside[processBuffer_sampleSize];
int32_t processBufferOutside[processBuffer_sampleSize];
const size_t processBufferShift_byteSize = (processBuffer_sampleSize / 2) * sizeof(int32_t); // half buffer, byte shift

float timeOffsetDelta = ((float)processBuffer_sampleSize / 2) / DOWNSAMPLED_FREQ; // 0.05sec
pcm_wav_header_t wave_header;

void processTestFiles();

#ifdef TESTING  // ---------- read from wav files
float startTimeOffset = 0;
uint32_t testDeltaT_counter = startTimeOffset / timeOffsetDelta;
const char WAVInsideFilename[] = "/RecInside.wav"; //The name of the test input sound
const char WAVOutsideFilename[] = "/RecOutside.wav"; //The name of the test input sound
File wavInsideFile;
File wavOutsideFile;

#else  //  -------------- read from mics,  no input wav files
const char *WAVInsideFilename = NULL;
const char *WAVOutsideFilename = NULL;
#endif  // #ifdef TESTING

// record 0,05ses at a time from Mics
uint32_t counter = 0;


bool SD_started = false;
float timeOffset = 0;

// Create an instance of the I2SClass
I2SClass i2s;


// updates inside and outside process buffers (globals) from byte arrays of 0.05 samples
void updateProcessBuffers(char *rec_inside, char *rec_outside, size_t rec_byteSize);


// updates inside and outside process buffers (globals) from byte arrays of 0.05 samples
void updateProcessBuffers(char *rec_inside, char *rec_outside, size_t byteSize) {
  if (byteSize != processBufferShift_byteSize) {
  }
  // valid data now
  memmove(processBufferInside, ((char*)processBufferInside) + processBufferShift_byteSize, processBufferShift_byteSize); // note void * so use byte sizes
  memcpy(((char*)processBufferInside) + processBufferShift_byteSize, rec_inside, processBufferShift_byteSize);
  memmove(processBufferOutside, ((char*)processBufferOutside) + processBufferShift_byteSize, processBufferShift_byteSize); // note void * so use byte sizes
  memcpy(((char*)processBufferOutside) + processBufferShift_byteSize, rec_outside, processBufferShift_byteSize);
}


// returns false if fails
bool recreateMagFile(const char* magFileName) {
  if (!magFileName) {
    return false;
  }
  bool rtn = true;
  File magFile;
  SD.remove(magFileName);
  //Open the magFile, create one if there is no magFile
  magFile = SD.open(magFileName, FILE_WRITE);
  if (!magFile) {
    rtn = false;
    Serial.print("Failed to "); Serial.print(magFileName); Serial.print(" for appending"); Serial.println();
  } else {
    magFile.print("CounterOffset(0.05s),"); magFile.print("time(s),"); magFile.print("Rain Pulse,"); magFile.print("Ratio In/Out,");
    magFile.print("Inside maxMag,"); magFile.print("Inside Freq,");
    magFile.print("Outside maxMag,"); magFile.print("Outside Freq");
    magFile.println();
    magFile.close();
  }
  return rtn;
}

void startTelentServer() {
  //setAsyncDebugPtr(&Serial); // enabled debug output for HP_AsyncTCP (slows down loop())
  if (initAsyncServer(portNo)) {
    Serial.print("HS_AsyncTCP started on port:"); Serial.println(portNo);
  } else {
    Serial.println("HS_AsyncTCP failed to started");
  }
  asyncConnectionTimeout(CONNECTION_TIMEOUT_MS); // enable this for pfodApp since it sends keepalives every 5sec
}


void TaskI2SRead(void *pvParameters);
TaskHandle_t i2s_read_task_handle;  // You can (don't have to) use this to be able to manipulate a task from somewhere else.

void startUpMic() {
  // Set up the pins used for audio input
  i2s.setPins(I2S_SCK, I2S_WS, -1, I2S_DIN);
  // Initialize the I2S bus in standard mode
  if (!i2s.begin(I2S_MODE_STD, SAMPLE_FREQ, DATA_BIT_WIDTH, I2S_SLOT_MODE_STEREO)) {
    while (1) {
      Serial.println("Failed to initialize I2S bus!");
      delay(5000);
    }
  }
  for (size_t i = 0; i < deltaT_counter_ignore; i++) {
    i2s.readBytes(rec_buffer_1, rec_buffer_byteSize); // discard first 0.5sec of mics output
  }

  // This variant of task creation can also specify on which core it will be run (only relevant for multi-core ESPs)
  xTaskCreatePinnedToCore(
    TaskI2SRead, "I2S Read", 1024  // Stack size
    , NULL  // When no parameter is used, simply pass NULL
    , 1  // Priority
    , &i2s_read_task_handle  // With task handle we will be able to manipulate with this task.
    , ARDUINO_RUNNING_CORE  // Core on which the task will run
  );
}


volatile bool read_1_Complete = false;
volatile bool read_2_Complete = false;
volatile bool read_1_Processed = true;
volatile bool read_2_Processed = true;
volatile size_t read_size_1 = 0;
volatile size_t read_size_2 = 0;
volatile bool read_1_delayed = false;
volatile bool read_2_delayed = false;
volatile bool taskReleased = false;

void releaseTask() {
  if (!taskReleased) {
    loopTimer.clear();
    taskReleased = true;
  }
}

void TaskI2SRead(void *pvParameters) {  // This is a task.
  (void)pvParameters;
  /*
    Read the mic
  */
  while (!taskReleased) { // wait for loop to run
    delay(1);
  }
  while (1) {
    if (!read_1_Processed) {
      read_1_delayed = true;
    }
    while (!read_1_Processed) {
      delay(1);
    }
    read_1_Processed = false;
    read_size_1 = i2s.readBytes(rec_buffer_1, rec_buffer_byteSize);
    read_1_Complete = true;
    if (!read_2_Processed) {
      read_2_delayed = true;
    }
    while (!read_2_Processed) {
      delay(1);
    }
    read_2_Processed = false;
    read_size_2 = i2s.readBytes(rec_buffer_2, rec_buffer_byteSize);
    read_2_Complete = true;
  }
}

// note insideBuffer and outsideBuffer must be at least len/2 long
// returns left righ buffer size in bytes
size_t splitStereoToMono(const char* recBuffer, size_t len, const char* insideBuffer, const char* outsideBuffer) {
  int32_t *recBuffer_32 = (int32_t*)recBuffer;
  int32_t *insideBuffer_32 = (int32_t*)insideBuffer;
  int32_t *outsideBuffer_32 = (int32_t*)outsideBuffer;
  size_t len_32 = len / bytesPerSample;
  size_t idx = 0;
  for (size_t i = 0; i < len_32 / 2; i++) { // pairs of 32bit words
    insideBuffer_32[i] = recBuffer_32[idx++];
    outsideBuffer_32[i] = recBuffer_32[idx++];
  }
  return len / 2;
}

// NOTE the global isRaining_v holds the is raining current state
uint32_t deltaT_counter = 0; // counter for 0.05 sec samples
void processMics() {
  if (read_1_Complete) {
    read_1_Complete = false;
    // ignore delays when reading from SD card test file
    if ((read_1_delayed) || (read_size_1 != rec_buffer_byteSize)) {
      sfLogLine.clear();
      if (read_1_delayed) {
        sfLogLine.print(deltaT_counter); sfLogLine.println(",1 Delayed");
      } else {
        unsigned long read_size = read_size_1;
        sfLogLine.print(deltaT_counter); sfLogLine.print(",1 short read : "); sfLogLine.print(read_size);
      }
      LogFile::write(sfLogLine);
      Serial.println(sfLogLine);
      read_1_delayed = false;
    }
    splitStereoToMono(rec_buffer_1, rec_buffer_byteSize, inside_buffer, outside_buffer);
    Goertzel::downSample32bitBy4((int32_t*)inside_buffer, mono_buffer_sampleSize);
    Goertzel::downSample32bitBy4((int32_t*)outside_buffer, mono_buffer_sampleSize);
    updateProcessBuffers(inside_buffer, outside_buffer, processBufferShift_byteSize);  // shift into process buffers
    processBuffers(deltaT_counter++, processBuffer_sampleSize, processBufferInside, processBufferOutside, DOWNSAMPLED_FREQ, filenameMag);
    read_size_1 = 0;
    read_1_Processed = true;
  }
  if (read_2_Complete) {
    read_2_Complete = false;
    // ignore delays when reading from SD card test file
    if ((read_2_delayed) || (read_size_2 != rec_buffer_byteSize)) {
      sfLogLine.clear();
      if (read_2_delayed) {
        sfLogLine.print(deltaT_counter); sfLogLine.println(",2 Delayed");
      } else {
        unsigned long read_size = read_size_2;
        sfLogLine.print(deltaT_counter); sfLogLine.print(",2 short read : "); sfLogLine.print(read_size);
      }
      LogFile::write(sfLogLine);
      Serial.println(sfLogLine);
      read_2_delayed = false;
    }
    splitStereoToMono(rec_buffer_2, rec_buffer_byteSize, inside_buffer, outside_buffer);
    Goertzel::downSample32bitBy4((int32_t*)inside_buffer, mono_buffer_sampleSize);
    Goertzel::downSample32bitBy4((int32_t*)outside_buffer, mono_buffer_sampleSize);
    updateProcessBuffers(inside_buffer, outside_buffer, processBufferShift_byteSize);  // shift into process buffers
    processBuffers(deltaT_counter++, processBuffer_sampleSize, processBufferInside, processBufferOutside, DOWNSAMPLED_FREQ, filenameMag);
    read_size_2 = 0;
    read_2_Processed = true;
  }
}

//#define PRINT_SIZES

void setup() {
  // Initialize the serial port
  Serial.begin(115200);
#ifdef DEBUG
  for (int i = 10; i > 0; i--) {
    delay(500); // wait 5sec to allow use to open plotter/monitor
  }
  bufferedOut.connect(Serial);
  debugPtr = &bufferedOut; // &Serial;
#endif

#ifndef TESTING
  //setESPAutoWiFiConfigDebugOut(Serial); // not this prints out network password!!
  if (ESPAutoWiFiConfigSetup(-5, true, 0)) { // Firebeetle ESP32-E has RGB led on GPIO5 so use -5 for led
    return;
  }
#endif

#ifdef PRINT_SIZES
  Serial.print(" rec_buffer_sampleSize:"); Serial.println(rec_buffer_sampleSize);
  Serial.print(" rec_buffer_byteSize:"); Serial.println(rec_buffer_byteSize);
  Serial.print(" mono_buffer_byteSize:"); Serial.println(mono_buffer_byteSize);
  Serial.print(" mono_buffer_sampleSize:"); Serial.println(mono_buffer_sampleSize);
  Serial.print(" processBuffer_sampleSize:"); Serial.println(processBuffer_sampleSize);
  Serial.print(" processBufferShift_byteSize:"); Serial.println(processBufferShift_byteSize);
  Serial.print(" timeOffsetDelta:"); Serial.println(timeOffsetDelta, 4);
#endif

  if (DATA_BIT_WIDTH != I2S_DATA_BIT_WIDTH_32BIT) {
    Serial.println("ERROR: Only works with 32bit data");
    while (1) {
      delay(3000);
    }
  }
  if (WAVInsideFilename || WAVOutsideFilename || filenameMag || logFilename) {
    if (!SD.begin(SD_CS_pin)) {
      Serial.println("SD begin failed");  // allow to run without an SD card
      SD_started = false;
    } else {
      SD_started = true;
    }
    while (!SD.begin(SD_CS_pin)) {
      Serial.print(".");
      delay(500);
    }
  }

#ifdef LOG_FILE
  LogFile::createLogFile(logFilename); // handles null
#endif

  recreateMagFile(filenameMag); // handles null filenameMag

  setProcessBuffersDebugOut(debugPtr); // outputs rain pulses to Serial if DEBUG

  //Goertzel::setDebugOut(debugPtr); dumps FT for selected freq range e.g. 660Hz to 770Hz
  initProcessBuffers(printScaleFactor, logRainPulses);
  memset(processBufferInside, 0, processBuffer_sampleSize * sizeof(int32_t)); // clear process buffer
  memset(processBufferOutside, 0, processBuffer_sampleSize * sizeof(int32_t)); // clear process buffer
  // NOTE will ignore first 5sec after start to let mic settle.

#ifdef TESTING
  Serial.println();
#if TESTING == BACK_TESTING
  Serial.println(" =====  Testing wav files for rain detection using current settings =========");
#else
  Serial.println(" =====  Dumping ALL ratios and magnitueds for these files =========");
#endif
  Serial.print("Processing previously recorded wav files, "); Serial.print(WAVInsideFilename); Serial.print(" and "); Serial.println(WAVOutsideFilename);
  setTestWAVinputDebugOut(&Serial);
  if (!initWAVinput(WAVInsideFilename, WAVOutsideFilename, wavInsideFile, wavOutsideFile, wave_header)) {
    while (1) {
      Serial.println("ERROR: initWAVinput failed");
      delay(5000);
    }
  }
#else  // normal running
  startTelentServer();
  startUpMic(); // also starts task, but does not release it to run yet
#endif
}

//#define DEBUG_LOOP_TIME

void loop() {
#ifdef DEBUG
  bufferedOut.nextByteOut(); // push bytes if it will not block // used by DEBUG
#endif
#ifdef DEBUG_LOOP_TIME
  loopTimer.check(Serial);  // max time here should be < 0.5sec to process and save 2 buffers
#endif

#ifdef TESTING
  processTestFiles();  // process pre-recorded wave forms
#else
// normal run
  if (ESPAutoWiFiConfigLoop()) {  // handle auto configure/connect
    return;
  }
  releaseTask(); // does nothing if TESTING
  processMics(); // process live data from stero mics
#endif
}

#ifdef TESTING
// =============  method to process pre-recorded downsampled mono files ===================
void processTestFiles() {
  size_t readSize = readWAVsec(wavInsideFile, wavOutsideFile, wave_header, testDeltaT_counter * timeOffsetDelta, inside_buffer,  outside_buffer, processBufferShift_byteSize);

  if (readSize != processBufferShift_byteSize) {
    while (1) {
      Serial.println("End of file");
      delay(5000);
    }
  }
  // process files
  updateProcessBuffers(inside_buffer, outside_buffer, readSize);
  // shift into process buffers
  processBuffers(testDeltaT_counter, processBuffer_sampleSize, processBufferInside, processBufferOutside, DOWNSAMPLED_FREQ, filenameMag);
  if (testDeltaT_counter % 200 == 0) {
    Serial.print(testDeltaT_counter * timeOffsetDelta / 60); Serial.println(" mins");
  }
  testDeltaT_counter++; // move on to next 0.05 section
}
#endif
