// ProcessBuffers.cpp
/*
 * (c)2024 Forward Computing and Control Pty. Ltd.
 * NSW Australia, www.forward.com.au
 * This code is not warranted to be fit for any purpose. You may only use it at your own risk.
 * This generated code may be freely used for both private and commercial use
 * provided this copyright is maintained.
 */

#include "ProcessBuffers.h"
#include "FS.h"
#include "SD.h"
#include "SafeString.h"
#include "Goertzel.h"
#include "LogFile.h"
#include "VolatileVars.h"


extern float timeOffsetDelta;

// check magnitude between these values
const int startFreq_H = 660;
const int endFreq_H = 770;

// as shown in the mag.txt file
const float MIN_HIGH_MAG = 4000; // was 5000; // times 1e5,  ignore if 700hz range magnitude is less than this number times 1e5
const float MIN_RATIO = 8;// was 20; // ignore ratio of inside/outside less than this

// set in initProcessBuffers()
static float FLOAT_MIN_HIGH_MAG = 0; // = MIN_HIGH_MAG/1e-5,  NOTE: mag log has numbers scaled by 1e-5 so the real MIN_HIGH_MAG is this value

// look for 1.5sec of pulses with less than 0.5 sec gap between each of them => is raining
static const uint32_t noRainStartDeltaTCount = 0.5 / timeOffsetDelta + 0.5; // == 0.5/0.05
static const uint32_t startRainDeltaTCount = 1.5 / timeOffsetDelta + 0.5; // == 1.5/0.05

// if have >= 8 pulses in 10sec => is raining
static const uint32_t pulseDeltaTCountUpdate = 0.25 / timeOffsetDelta + 0.5; // update pulse count each 0.25sec
static const size_t noOfPulseUpdatePoints = 10 * 4 ; // 40 x 0.25 => 10sec sum;
static const uint32_t MIN_CUMULATIVE_PULSES = 8; // 8 pulses in the last 10sec or less

// if not pulses for 10sec => rain stopped.
static const uint32_t rainStoppedDeltaTCount = 10.0 / timeOffsetDelta + 0.5; // == 10/0.05

static int data[noOfPulseUpdatePoints];
static long int sum = 0;
static int average = 0;
static size_t idx = 0;
static Print* debugPtr = NULL;


static bool logRainPulses = false;

static bool printHeader = true; // only on the first call

void setProcessBuffersDebugOut(Print *_debugPtr) {
  debugPtr = _debugPtr;
}

static const size_t maxLogLineLen = 400;
static char logLine[maxLogLineLen];

bool processRainPulse(bool isRainPulse, uint32_t offsetCounter);

void initRollingSumInt() {
  for (size_t i = 0; i < noOfPulseUpdatePoints; i++) {
    data[i] = 0;
  }
}

size_t getnoOfPulseUpdatePoints() {
  return noOfPulseUpdatePoints;
}

long int updateSum(int newData) {
  sum -= data[idx];
  data[idx] = newData;
  idx++;
  if (idx >= noOfPulseUpdatePoints) {
    idx = 0;
  }
  sum += newData;
  return sum;
}


void initProcessBuffers(float printScaleFactor, bool _logRainPulses) {
  logRainPulses = _logRainPulses;
  Goertzel::setPrintScaleFactor(printScaleFactor); // divide output prints by this number as print(float) prints ovf for numbers > 4294967040
  FLOAT_MIN_HIGH_MAG = MIN_HIGH_MAG / printScaleFactor;
  initRollingSumInt();
  cSFPS(sfLogLine, logLine, maxLogLineLen);
  sfLogLine.clear();
  sfLogLine.println(" Pulse Detection Settings:-");
  sfLogLine.print(" deltaT:"); sfLogLine.println(timeOffsetDelta,2);
  sfLogLine.print(" minRatio:"); sfLogLine.println(MIN_RATIO);
  sfLogLine.print(" minMag inside:"); sfLogLine.println(MIN_HIGH_MAG);
  sfLogLine.print(" noRainStartDeltaTCount:"); sfLogLine.println(noRainStartDeltaTCount);
  sfLogLine.print(" startRainDeltaTCount:"); sfLogLine.println(startRainDeltaTCount);
  sfLogLine.print(" rainStoppedDeltaTCount:"); sfLogLine.println(rainStoppedDeltaTCount);
  sfLogLine.print(" pulseDeltaTCountUpdate:"); sfLogLine.println(pulseDeltaTCountUpdate);
  sfLogLine.print(" noOfPulseUpdatePoints:"); sfLogLine.println(noOfPulseUpdatePoints);
  sfLogLine.print(" MIN_CUMULATIVE_PULSES:"); sfLogLine.println(MIN_CUMULATIVE_PULSES);
  sfLogLine.println();
  if (logRainPulses) {
    LogFile::write(sfLogLine);
  }
  if (debugPtr) {
    debugPtr->print(sfLogLine);
    debugPtr->flush(); // push bufferedOut to Serial 
  }
  sfLogLine.clear();
}


// performs FT and returns true if rain detected
bool processBuffers(uint32_t offsetCounter, size_t sampleSize, int32_t* insideSamples, int32_t* outsideSamples, float sampleFreq, const char* magFileOut) {
  cSFPS(sfLogLine, logLine, maxLogLineLen);
  sfLogLine.clear();

  if (printHeader) {
    printHeader = false;
    sfLogLine.println("time,ratio,Inside_Max,Inside_Hz,Outside_Max,Outside_Hz,isRaining");
    if (logRainPulses) {
      LogFile::write(sfLogLine);
    }
    if (debugPtr) {
      debugPtr->print(sfLogLine);
    }
    sfLogLine.clear();
  }
  bool isRainPulse = false; // no rain pulse
  File magFile;
  if (magFileOut) {
    magFile = SD.open(magFileOut, FILE_APPEND);
    if (!magFile) {
      if (debugPtr) {
        debugPtr->println("Failed to open file for appending");
      }
    }
  }
  float timeOffset = offsetCounter * timeOffsetDelta;
  float maxFreq_Inside = 0;
  float maxFreq_Outside = 0;
  float maxMagnitude_Inside = Goertzel::maxMag32bitQuarterSteps(startFreq_H, endFreq_H, insideSamples, sampleSize, sampleFreq, &maxFreq_Inside);
  float maxMagnitude_Outside = Goertzel::maxMag32bitQuarterSteps(startFreq_H, endFreq_H, outsideSamples, sampleSize, sampleFreq, &maxFreq_Outside);
  float ratio = 0;
  if ((maxMagnitude_Outside > 0) && (maxMagnitude_Inside > 0)) {
    ratio = maxMagnitude_Inside / maxMagnitude_Outside;
  }
  if  ((ratio > MIN_RATIO) && (maxMagnitude_Inside > FLOAT_MIN_HIGH_MAG)) {  // i.e. MIN_HIGH_MAG / Goertzel::getPrintScaleFactor()
    isRainPulse = true;
  }
  // update telnet values
  if (ratio > maxRatio_v) {
    maxRatio_v = ratio;
    maxRatioMagnitude_v = maxMagnitude_Inside  * Goertzel::getPrintScaleFactor();
  }
  if (isRainPulse) {
    sfLogLine.print(timeOffset, 3); sfLogLine.print(','); sfLogLine.print(ratio, 4); sfLogLine.print(',');
    sfLogLine.print(maxMagnitude_Inside * Goertzel::getPrintScaleFactor()); sfLogLine.print(','); sfLogLine.print(maxFreq_Inside, 2); sfLogLine.print(',');
    sfLogLine.print(maxMagnitude_Outside * Goertzel::getPrintScaleFactor()); sfLogLine.print(','); sfLogLine.print(maxFreq_Outside, 2);
  } else {
    sfLogLine.print(timeOffset, 3); sfLogLine.print(',');  sfLogLine.print(',');
    sfLogLine.print(',');  sfLogLine.print(',');
    sfLogLine.print(',');
  }

  if (magFile) {
    magFile.print(offsetCounter); magFile.print(',');
    magFile.print(timeOffset); magFile.print(',');
    if (isRainPulse) {
      magFile.print("1,");
    } else {
      magFile.print("0,");
    }
    magFile.print(ratio, 4); magFile.print(',');
    magFile.print(maxMagnitude_Inside * Goertzel::getPrintScaleFactor()); magFile.print(','); magFile.print(maxFreq_Inside, 2); magFile.print(',');
    magFile.print(maxMagnitude_Outside * Goertzel::getPrintScaleFactor()); magFile.print(','); magFile.print(maxFreq_Outside, 2); magFile.println();
    magFile.close();
  }
  return processRainPulse(isRainPulse, offsetCounter);
}

// if no rain pulse for 2sec rain stopped
// to start rain need at least 1 rain pulse every 0.25sec for 2 sec

uint32_t lastPulseCounter = 0;
uint32_t startRainCounter = 0;
uint32_t pulseCountPerUpdate = 0;

bool processRainPulse(bool isRainPulse, uint32_t offsetCounter) {
  deltaT_counter_v = offsetCounter; // for WiFi logging of pulses
  cSFPS(sfLogLine, logLine, maxLogLineLen);

  static bool isRaining = false;
  static char state = 'N';
  bool logLineWritten = false;


  if (isRainPulse) {
    rainPulses_v = rainPulses_v + 1; // only updated here, only read in WiFiDataHandling
    if (!isRaining) {
      if (startRainCounter == 0) {
        state = 'S';
        sfLogLine.print(",0");
        sfLogLine.println();
        if (logRainPulses) {
          LogFile::write(sfLogLine);
        }
        if (debugPtr) {
          debugPtr->print(sfLogLine);
        }
        startRainCounter = offsetCounter; // start timer again
      } else { // check for time out no rain
        if ((offsetCounter - lastPulseCounter) > noRainStartDeltaTCount) {
          startRainCounter = offsetCounter;
          state = 'S';
          sfLogLine.print(",0");
          sfLogLine.print(",  >> start rain timed out");
          sfLogLine.println();
          if (logRainPulses) {
            LogFile::write(sfLogLine);
          }
          if (debugPtr) {
            debugPtr->print(sfLogLine);
          }
        } else {
          state = '.';
          if ((offsetCounter - startRainCounter) > startRainDeltaTCount) {
            state = 'R';
            isRaining = true;
            isRaining_v = true;
            isRaining_counter_v = offsetCounter;
            startRainCounter == 0;
            // end of rain
            cSF(sfCopy, maxLogLineLen);
            sfCopy = sfLogLine;
            sfCopy.print(",0");
            sfCopy.println();
            if (logRainPulses) {
              LogFile::write(sfCopy);
            }
            if (debugPtr) {
              debugPtr->print(sfCopy);
            }
            //
            sfLogLine.print(",1");
            sfLogLine.print(",  >> Rain Started");
            sfLogLine.println();
            if (logRainPulses) {
              LogFile::write(sfLogLine);
            }
            if (debugPtr) {
              debugPtr->print(sfLogLine);
            }
          } else {
            sfLogLine.print(",0");
            sfLogLine.println();
            if (logRainPulses) {
              LogFile::write(sfLogLine);
            }
            if (debugPtr) {
              debugPtr->print(sfLogLine);
            }
          }
        }
      }
    } else {  // raining
      // print magnitude and ratio after start Raining
      sfLogLine.print(",1");
      sfLogLine.println();
      if (logRainPulses) {
        LogFile::write(sfLogLine);
      }
      if (debugPtr) {
        debugPtr->print(sfLogLine);
      }
    }
    lastPulseCounter = offsetCounter;

  } else { // no rain pulse
    if (!isRaining) {
      if ((state != 'N') && ((offsetCounter - lastPulseCounter) > noRainStartDeltaTCount)) {
        state = 'N';
        startRainCounter = 0;
        sfLogLine.print(",0");
        sfLogLine.print(",  >> start rain timed out - no pulse");
        sfLogLine.println();
        if (logRainPulses) {
          LogFile::write(sfLogLine);
        }
        if (debugPtr) {
          debugPtr->print(sfLogLine);
        }
      }
    } else { // isRaining
      // if have not had a rain pulse for 10 sec => not Raining
      if (((offsetCounter - lastPulseCounter) > rainStoppedDeltaTCount)) {
        state = 'N';
        isRaining = false;
        isRaining_v = false;
        isRaining_counter_v = offsetCounter;
        startRainCounter = 0;
        // end of rain
        cSF(sfCopy, maxLogLineLen);
        sfCopy = sfLogLine;
        sfCopy.print(",1");
        sfCopy.println();
        if (logRainPulses) {
          LogFile::write(sfCopy);
        }

        if (debugPtr) {
          debugPtr->print(sfCopy);
        }
        //
        sfLogLine.print(",0");
        sfLogLine.println(",Rain Stopped");
        if (logRainPulses) {
          LogFile::write(sfLogLine);
        }
        if (debugPtr) {
          debugPtr->print(sfLogLine);
        }
      }
    }
  }

  if (isRainPulse) {
    pulseCountPerUpdate++;
  }

  if ((offsetCounter % pulseDeltaTCountUpdate) == 0) {
    long int sum = updateSum(pulseCountPerUpdate);
    pulseCountPerUpdate = 0;
    if (sum >= MIN_CUMULATIVE_PULSES) { // also handles 8 in one period
      if (!isRaining) {
        state = 'R';
        isRaining = true;
        isRaining_v = true;
        isRaining_counter_v = offsetCounter;
        startRainCounter == 0;
        sfLogLine.clear();
        sfLogLine.print(offsetCounter * timeOffsetDelta , 3); sfLogLine.print(',');  sfLogLine.print(',');
        sfLogLine.print(',');  sfLogLine.print(',');
        sfLogLine.print(',');
        // end of rain
        cSF(sfCopy, maxLogLineLen);
        sfCopy = sfLogLine;
        sfCopy.print(",0");
        sfCopy.println();
        if (logRainPulses) {
          LogFile::write(sfCopy);
        }
        if (debugPtr) {
          debugPtr->print(sfCopy);
        }


        sfLogLine.print(",1");
        sfLogLine.print(",  >> Rain Started - cumulative pulses");
        sfLogLine.println();
        if (logRainPulses) {
          LogFile::write(sfLogLine);
        }
        if (debugPtr) {
          debugPtr->print(sfLogLine);
        }
      }
    }
  }

  return isRaining;
}
